#!/usr/bin/env python3

import sys
import re
import os
import asyncio

import irc3
import irc3.utils

from irc3.plugins.command import command

from .rcon import RconConnection

try:
    from systemd import journal
except ImportError:
    journal = None

ONLINE_RE = re.compile(r'\s*(.*?)\s+\(online\)')
LOG_PATTERN = r'\s*(?P<time>[\d.]+) (?P<level>Info|Verbose|Warning|Error) '
JOIN_PART_RE = re.compile(LOG_PATTERN + r'[^ ]+ MapTick\(\d+\) processed Player(?P<action>Leave|Join)Game peerID\((?P<peer_id>\d+)\).*')
USERNAME_RE = re.compile(LOG_PATTERN + r'[^ ]+ Received peer info for peer\((?P<peer_id>\d+)\) username\((?P<username>[^)]+)\).*')
CHAT_RE = re.compile(r'(?P<username>[^: ]+): (?P<message>.*)')


class SystemdJournalLogReader:
    def __init__(self, loop, callback, unit='factorio.service', **kwargs):
        if journal is None:
            raise ImportError('Please install the systemd python module')

        self.loop = loop
        self.callback = callback
        self.reader = journal.Reader()
        self.reader.add_match(_SYSTEMD_UNIT=unit)
        self.reader.seek_tail()

        # seek_tail() still leaves a few messages at the end
        # so we have to consume them first
        list(self.reader)

        self.fd = self.reader.fileno()
        self.loop.add_reader(self.fd, self.on_fd_ready)

    def __del__(self):
        self.loop.remove_reader(self.fd)

    def on_fd_ready(self):
        for entry in self.reader:
            if not entry:  # empty dict = no more entries
                break
            self.callback(entry['MESSAGE'])


class StdinLogReader:
    def __init__(self, loop, callback, **kwargs):
        self.loop = loop
        self.callback = callback
        self.task = loop.create_task(self.log_read())

        # If stdin is a file, seek to its end
        if sys.stdin.seekable():
            sys.stdin.seek(0, os.SEEK_END)

    def __del__(self):
        self.task.cancel()

    async def log_read(self):
        while True:
            line = await self.loop.run_in_executor(
                None,
                sys.stdin.readline
            )
            if not line:  # EOF reached
                if not sys.stdin.seekable():
                    break
                await asyncio.sleep(0.2, loop=self.loop)
                continue
            line = line.strip('\n')
            self.callback(line)

READERS = {
    'stdin': StdinLogReader,
    'systemd': SystemdJournalLogReader,
}


@irc3.plugin
class FactoIRC:
    requires = ['irc3.plugins.command']

    def __init__(self, bot):
        self.bot = bot
        self.config = bot.config.get(self.__class__.__module__, {})
        autojoins = self.bot.config.get('autojoins')
        self.channels = irc3.utils.as_list(
            self.config.get('channels', autojoins)
        )
        self.reader = None
        self.peer_names = {}

    @irc3.event(irc3.rfc.JOIN)
    def on_join(self, mask, channel, **kwargs):
        if mask.nick == self.bot.nick:
            reader_class = READERS[self.config.get('method', 'stdin')]
            self.reader = reader_class(self.bot.loop,
                                       self.log_line,
                                       **self.config)

    @irc3.event(irc3.rfc.PRIVMSG)
    async def on_privmsg(self, mask, target, data, **kwargs):
        if target not in self.channels:
            return
        if data.startswith(self.bot.config.get('cmd', '!')):
            return

        await self.do_rcon('(irc) %s: %s' % (mask.nick, data))

    def broadcast(self, msg):
        for channel in self.channels:
            self.bot.privmsg(channel, '(factorio) %s' % msg)

    def log_line(self, line):
        for pattern in (CHAT_RE, JOIN_PART_RE, USERNAME_RE):
            m = pattern.match(line)
            if not m:
                continue

            if pattern == USERNAME_RE:
                peer_id = m.group('peer_id')
                username = m.group('username')
                self.peer_names[peer_id] = username

            elif pattern == JOIN_PART_RE:
                action = m.group('action')
                peer_id = m.group('peer_id')
                try:
                    username = self.peer_names[peer_id]
                except KeyError:
                    continue

                if action == 'Join':
                    msg = '%s joined the game'
                elif action == 'Leave':
                    msg = '%s left the game'

                self.broadcast(msg % username)

            elif pattern == CHAT_RE:
                username = m.group('username')
                if username == '<server>':
                    continue
                message = m.group('message')
                self.broadcast('%s: %s' % (username, message))

    async def do_rcon(self, text):
        host = self.config.get('rcon_host', 'localhost')
        port = self.config.get('rcon_port', '27015')
        password = self.config.get('rcon_password', '')

        conn = RconConnection(host, int(port), password, loop=self.bot.loop)
        result = (await conn.exec_command(text)).splitlines()
        return result

    @command(permission='rcon')
    async def rcon(self, mask, target, args):
        '''
            Execute an RCON command

            %%rcon <command>...
        '''
        cmd = ' '.join(args['<command>'])
        result = await self.do_rcon(cmd)
        return result

    @command(permission='players')
    async def players(self, mask, target, args):
        '''
            Show connected players.

            %%players
        '''
        players = await self.do_rcon('/players')
        players = [m.group(1)
                   for m in map(ONLINE_RE.match, players)
                   if m]
        if players:
            return 'Connected players (%d): %s' % (
                    len(players), ', '.join(players))
        else:
            return 'No one is connected'



# Copyright (C) 2016 Mickaël Thomas
# Copyright (C) 2013 Peter Rowlands

# From pysrcds project: https://github.com/pmrowla/pysrcds
# Converted to python3 and asyncio.
# Made compatible to the Factorio RCON implementation.

"""Source server RCON communications module"""

import struct
import itertools
import asyncio


# Packet types
SERVERDATA_AUTH = 3
SERVERDATA_AUTH_RESPONSE = 2
SERVERDATA_EXECCOMMAND = 2
SERVERDATA_RESPONSE_VALUE = 0


class RconPacket(object):

    """RCON packet"""

    _struct = struct.Struct('<3i')

    def __init__(self, pkt_id=0, pkt_type=-1, body=b''):
        self.pkt_id = pkt_id
        self.pkt_type = pkt_type
        self.body = body

    def __bytes__(self):
        """Return the body string"""
        return self.body

    def __repr__(self):
        return "RconPacket(pkt_id=%d, pkt_type=%d, body=%r)" % (
            self.pkt_id, self.pkt_type, self.body)

    def size(self):
        """Return the pkt_size field for this packet"""
        return len(self.body) + 10

    def pack(self):
        """Return the packed version of the packet"""
        header = self._struct.pack(self.size(), self.pkt_id, self.pkt_type)
        return b'%s%s\x00\x00' % (header, self.body)


class RconConnection(object):

    """RCON client to server connection"""

    def __init__(self, server, port=27015, password='', loop=None,
                 encoding='utf-8'):
        self.server = server
        self.port = port
        self.encoding = encoding
        self.password = password
        self.loop = loop
        self.authenticated = False
        self.pkt_id = itertools.count(1)

    async def authenticate(self, password=None):
        """Authenticate with the server using the given password"""

        if password is None:
            password = self.password
        password = password.encode(self.encoding)

        self.rd, self.wr = await asyncio.open_connection(
            self.server, self.port, loop=self.loop
        )

        auth_pkt = RconPacket(next(self.pkt_id), SERVERDATA_AUTH, password)
        await self._send_pkt(auth_pkt)

        auth_resp = await self.read_response()

        if auth_resp.pkt_type != SERVERDATA_AUTH_RESPONSE:
            raise RconError('Received invalid auth response packet')

        if auth_resp.pkt_id == -1:
            raise RconAuthError('Bad password')

        self.authenticated = True

    async def exec_command(self, command, read_response=True):
        """
        Execute the given RCON command
        Return the response body
        """

        if not self.authenticated:
            await self.authenticate()

        cmd_pkt = RconPacket(next(self.pkt_id), SERVERDATA_EXECCOMMAND,
                             command.encode(self.encoding))

        await self._send_pkt(cmd_pkt)
        if read_response:
            resp = await self.read_response(cmd_pkt)
            return resp.body.decode(self.encoding)
        else:
            return cmd_pkt

    async def _send_pkt(self, pkt):
        """Send one RCON packet over the connection"""

        data = pkt.pack()
        self.wr.write(data)
        await self.wr.drain()

    async def _recv_pkt(self):
        """Read one RCON packet"""

        header = await self.rd.readexactly(struct.calcsize('<3i'))
        (pkt_size, pkt_id, pkt_type) = struct.unpack('<3i', header)
        body = await self.rd.readexactly(pkt_size - 8)

        # Strip the 2 trailing nulls from the body
        body = body.rstrip(b'\x00')

        return RconPacket(pkt_id, pkt_type, body)

    async def read_response(self, request=None):
        """Return the next response packet"""

        if request and not isinstance(request, RconPacket):
            raise TypeError('Expected RconPacket type for request')

        response = await self._recv_pkt()

        if (response.pkt_type != SERVERDATA_RESPONSE_VALUE and
                response.pkt_type != SERVERDATA_AUTH_RESPONSE):
            raise RconError('Recieved unexpected RCON packet type')

        if request and response.pkt_id != request.pkt_id:
            raise RconError('Response ID does not match request ID')

        return response


class RconError(Exception):
    """Generic RCON error"""
    pass


class RconAuthError(RconError):
    """Raised if an RCON Authentication error occurs"""


def main():
    import sys

    if len(sys.argv) < 5:
        print('Usage: %s <host> <port> <password> <command>...' % sys.argv[0])
        sys.exit(1)

    host, port, password, *cmd = sys.argv[1:]
    port = int(port)
    cmd = ' '.join(cmd)

    loop = asyncio.get_event_loop()
    conn = RconConnection(host, port, password)
    resp = loop.run_until_complete(conn.exec_command(cmd))
    print(resp, end='')

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

with open('README.rst', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='factoirc',

    version='0.2',

    description='Bidirectional IRC bridge for Factorio',
    long_description=long_description,

    url='https://github.com/mickael9/factoirc',

    author='Mickaël Thomas',
    author_email='mickael9@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.5',

        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Games/Entertainment',
    ],

    keywords='factorio irc',

    packages=find_packages(),

    install_requires=['irc3'],

    extras_require={
        'systemd': ['python-systemd'],
    },
)
